import yaml
import re
import copy
import math


class Unit(float):

    def __add__(self, other):
        return self.__class__(float.__add__(self, other))

    def __sub__(self, other):
        return self.__class__(float.__sub__(self, other))

    def __div__(self, other):
        return self.__class__(float.__div__(self, other))

    def __floordiv__(self, other):
        return self.__class__(float.__floordiv__(self, other))

    def __mul__(self, other):
        return self.__class__(float.__mul__(self, other))

    def __str__(self):
        return "{}({})".format(self.__class__.__name__, float(self))

    def __repr__(self):
        return str(self)

    def __neg__(self):
        return self.__class__(-float(self))

    def __abs__(self):
        return self.__class__(abs(float(self)))

class Distance(Unit):
    @classmethod
    def from_mm(cls, val):
        return cls(val / 1000)

    def as_mm(self):
        return float(self * 1000)

    def as_m(self):
        return float(self)



class Angle(Unit):
    @classmethod
    def from_degree(cls, val):
        return cls(math.radians(val))

    def as_degrees(self):
        return math.degrees(self)

    def as_radians(self):
        return float(self)

class Current(Unit):
    @classmethod
    def from_ma(cls, val):
        return cls(val/1000)

    def as_ma(cls, val):
        return float(val*1000)

class Torque(Unit):
    @classmethod
    def from_kg_cm(cls,val):
        return cls((9.8*val)/100)

    def as_n_m(self):
        return float(self)

class AngleVelocity(Unit):
    @classmethod
    def from_s_60_degree(cls, val):
        degrees_sec = 60/val

        return cls( math.radians(degrees_sec) )

    def as_radians_s(self):
        return float(self)

unit_config = {
            "mm": Distance.from_mm,
            "m": Distance,
            u"\xb0": Angle.from_degree,
            "rad": Angle,
            "r": Angle,
            "A": Current,
            "mA": Current.from_ma,
            "kg/cm": Torque.from_kg_cm,
            u"sec/60\xb0": AngleVelocity.from_s_60_degree
        }

config_re = re.compile(u"^\s*([-\d.]+)({})\s*$".format("|".join(unit_config.keys())))

def cfg_constructor(loader, node):
    value = loader.construct_scalar(node)
    res = re.match(config_re, value)

    val, unit = res.group(1), res.group(2)

    val = float(val)

    return unit_config[unit](val)

yaml.add_constructor(u'!cfg', cfg_constructor)
yaml.add_implicit_resolver(u'!cfg', config_re)

def _parse_val(val, top_cfg):

    res = re.match("^\s*%(.*)%\s*$", val)
    if res is None:
        return 0, 0, val

    expr = res.group(1)
    base_expr = expr
    vars = re.findall("[a-zA-Z][a-zA-Z0-9\._]+", expr)
    replaced_count = 0
    unreplaced_count = 0

    for var in vars:
        if var in ["max", "min", "abs", "Distance", "Angle"]:
            continue
        keys = var.split('.')
        value = top_cfg
        while len(keys) > 0:
            if isinstance(value, list):
                value = value[int(keys[0])]
            elif isinstance(value, dict):
                try:
                    value = value[keys[0]]
                except KeyError:
                    raise Exception(
                        "Using key for variable that does not exist ", keys[0],
                        " in expression: ", base_expr)
            keys = keys[1:]
        if isinstance(value, str) and re.match("^\s*%(.*)%\s*$", value):
            unreplaced_count += 1
            continue
        expr = expr.replace(var, str(value))
        replaced_count += 1

    if unreplaced_count > 0:
        expr = "%" + expr + "%"
        return replaced_count, unreplaced_count, expr

    try:
        print expr
        expr = eval(expr)
        print expr
    except:
        expr = "%" + expr + "%"

    return replaced_count, unreplaced_count, expr


# returns number of replaced values
def _parse_config_rec(cfg, top_cfg):

    keys = list()
    if isinstance(cfg, list):
        keys = range(len(cfg))

    elif isinstance(cfg, dict):
        keys = cfg.keys()

    replaced_big_count = 0
    unreplaced_big_count = 0
    for key in keys:
        if isinstance(cfg[key], (float, int, bool)):
            continue
        elif isinstance(cfg[key], str):
            replaced_count, unreplaced_count, res = _parse_val(
                cfg[key], top_cfg)
            replaced_big_count += replaced_count
            unreplaced_big_count += unreplaced_count
            cfg[key] = res
        else:
            replaced_count, unreplaced_count = _parse_config_rec(
                cfg[key], top_cfg)
            replaced_big_count += replaced_count
            unreplaced_big_count += unreplaced_count

    return replaced_big_count, unreplaced_big_count


def parse_config(filename):
    config = yaml.load(file(filename, "r"))

    while _parse_config_rec(config, config)[0] > 0:
        pass

    if _parse_config_rec(config, config)[1] > 0:
        raise Exception("Some values in config expressions were not replaced ",
                        config)


    return config
